/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm;

import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;

import javax.swing.Timer;
import javax.swing.event.ChangeEvent;

/**
 * Panel that provides a geometrical view with inner and outer stage animations.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RKAnimationPanel extends RKDrawing {
	
	// this drawing will need to be notified when animations end
	private RKDerivPanel rkdp;
	
	// min and max ordinate in use
	private double minY, maxY;
	private Timer t;
	// temporary point used for animations
	private Point2D anim;
	
	/**
	 * Create a new drawing for the given model.
	 * This drawing will notify the given RKDerivPanel when its animations end. 
	 */
	public RKAnimationPanel(RKModel rkm, RKDerivPanel rkdp) {
		super(rkm);
		this.rkdp = rkdp;
	}
	
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;
		
		if (getModel().getState() != State.EMPTY) {
			drawTitle(getModel().getODE().getSolutionDescription(), g2);
			drawIntervals(getModel().getStartPoint().getX(), getModel().getEndPoint().getX(), minY, maxY, g2);
			plotFunction(getModel().getStartPoint().getX(), getModel().getStartPoint().getY(), false, g2, getPlotColor(), new BasicStroke(1.5f));			
		}
		if (getModel().getState() == State.INNER) {
			// draw all but the current stage, without full information (just a slope is drawn)
			for (int i = 1; i < getModel().getInnerStagesDone(); i++)
				drawSample(i, false, g2);
			// if still animating, draw the animated point
			if (t.isRunning()) {
				g2.setPaint(getInnerStagesColor());				
				g2.fill(new Ellipse2D.Float(dx(anim.getX()) - 3, dy(anim.getY()) - 3, 6, 6));		
			}
			// else, draw this stage with full information
			else
				drawSample(getModel().getInnerStagesDone(), true, g2);
		}
		else if (getModel().getState() == State.OUTER) {
			// draw all inner stages without full information
			for (int i = 1; i <= getModel().getInnerStagesDone(); i++)
				drawSample(i, false, g2);
			// if still animating, draw just part of the approximated solution
			if (t.isRunning()) {
				g2.setPaint(getOuterStageColor());
				g2.draw(new Line2D.Float(dx(getModel().getStartPoint().getX()), dy(getModel().getStartPoint().getY()), dx(anim.getX()), dy(anim.getY())));
				g2.fill(new Ellipse2D.Float(dx(getModel().getStartPoint().getX()) - 3, dy(getModel().getStartPoint().getY()) - 3, 6, 6));
				g2.fill(new Ellipse2D.Float(dx(anim.getX()) - 3, dy(anim.getY()) - 3, 6, 6));
			}
			// else, draw it completely
			else
				drawSolution(g2);
		}
	}
	
	// draw an inner stage with or without full information
	private void drawSample(int sample, boolean full, Graphics2D g2) {
		Point2D[] sp = getModel().getSamplePoints();
		double[] s = getModel().getSamples();
		int i = sample - 1;
		
		// only plot the tangent solution if full info is requested
		if (full)
			plotFunction(sp[i].getX(), sp[i].getY(), false, g2, getSecondaryPlotColor(), new BasicStroke());
		
		tick(sp[i].getX(), sp[i].getY(), g2);
		
		g2.setPaint(getInnerStagesColor());
		
		// these formulas calculate the screen coordinates of the end points of the slope element
		// note that we want the length of the slope element to be 5% of the screen width
		double dx = (getWidth() * 0.05) / Math.sqrt(getResolutionX() * getResolutionX() + s[i] * s[i] * getResolutionY() * getResolutionY());
		float xs = dx(sp[i].getX() - dx);
		float ys = dy(sp[i].getY() - s[i] * dx);
		float xf = dx(sp[i].getX() + dx);
		float yf = dy(sp[i].getY() + s[i] * dx);
		g2.draw(new Line2D.Float(xs, ys, xf, yf));
		
		// fill the oval when providing full information, otherwise just draw an outline
		if (full)
			g2.fill(new Ellipse2D.Float(dx(sp[i].getX()) - 3, dy(sp[i].getY()) - 3, 6, 6));
		else
			g2.draw(new Ellipse2D.Float(dx(sp[i].getX()) - 3, dy(sp[i].getY()) - 3, 6, 6));
	}
	
	// draw the outer stage
	private void drawSolution(Graphics2D g2) {
		// tick what hasn't been ticked yet
		tickY(getModel().getStartPoint().getY(), g2, getCoordinateColor());
		tickY(getModel().getEndPoint().getY(), g2, getOuterStageColor());
		
		// draw the approximated solution as a line with ovals at the end
		g2.setPaint(getOuterStageColor());
		g2.draw(new Line2D.Float(dx(getModel().getStartPoint().getX()), dy(getModel().getStartPoint().getY()), dx(getModel().getEndPoint().getX()), dy(getModel().getEndPoint().getY())));	
		g2.fill(new Ellipse2D.Float(dx(getModel().getStartPoint().getX()) - 3, dy(getModel().getStartPoint().getY()) - 3, 6, 6));
		g2.fill(new Ellipse2D.Float(dx(getModel().getEndPoint().getX()) - 3, dy(getModel().getEndPoint().getY()) - 3, 6, 6));
		
		// indicate the error in the solution with a dotted line
		g2.setStroke(new BasicStroke(1f, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, new float[] {1, 2}, 0f));
		g2.draw(new Line2D.Float(dx(getModel().getEndPoint().getX()), dy(getModel().getODE().evalSol(getModel().getStartPoint().getX(), getModel().getStartPoint().getY(), getModel().getEndPoint().getX())), dx(getModel().getEndPoint().getX()), dy(getModel().getEndPoint().getY())));
		g2.setStroke(new BasicStroke());
		g2.draw(new Ellipse2D.Float(dx(getModel().getEndPoint().getX()) - 3, dy(getModel().getODE().evalSol(getModel().getStartPoint().getX(), getModel().getStartPoint().getY(), getModel().getEndPoint().getX())) - 3, 6, 6));
	}
	
	/**
	 * Update the view.
	 */
	public void stateChanged(ChangeEvent e) {
		// stop the animation in case it is still running
		if (t != null)
			t.stop();
		
		if (getModel().getState() == State.INITIALIZED) {
			// calculate minY and maxY (min/max of sample points and start/end point)
			minY = getModel().getStartPoint().getY();
			maxY = getModel().getStartPoint().getY();
			
			Point2D[] samplePoints = getModel().getSamplePoints();
			for (Point2D p : samplePoints)
				if (p.getY() < minY)
					minY = p.getY();
				else if (p.getY() > maxY)
					maxY = p.getY();
			
			if (getModel().getEndPoint().getY() < minY)
				minY = getModel().getEndPoint().getY();
			else if (getModel().getEndPoint().getY() > maxY)
				maxY = getModel().getEndPoint().getY();
			
			// we want an interval, even if minY = maxY
			if (minY == maxY) {
				maxY = minY + 0.5;
				minY -= 0.5;
			}

			moveToIntervals(getModel().getStartPoint().getX(), getModel().getEndPoint().getX(), minY, maxY);
		}
		else if (getModel().getState() == State.INNER) {
			anim = getModel().getStartPoint();
			// we set the timer to 50 FPS (a lower fps seems to cause cause choppy animation)
			t = new Timer(1000 / 50, new ActionListener() {
				
				private Point2D dest = getModel().getSamplePoints()[getModel().getInnerStagesDone() - 1];
				// find the total distance to travel in screen coordinates
				private double d = Math.sqrt((dx(dest.getX()) - dx(getModel().getStartPoint().getX())) * (dx(dest.getX()) - dx(getModel().getStartPoint().getX())) + (dy(dest.getY()) - dy(getModel().getStartPoint().getY())) * (dy(dest.getY()) - dy(getModel().getStartPoint().getY())));
				// we want to travel at a constant speed (half the width per second) and find the number of frames to animate
				private int nframes = (int) (d * 50 / (getWidth() / 2));
				private int frame = 1;
				
				public void actionPerformed(ActionEvent e) {
					if (frame <= nframes) {
						// move the point by 1/nframes of the total distance in true coordinates (so we end up at the right point, even if the view is moved/scaled during animation)
						anim.setLocation(anim.getX() + (dest.getX() - getModel().getStartPoint().getX()) / nframes, anim.getY() + (dest.getY() - getModel().getStartPoint().getY()) / nframes);
						frame++;
						repaint();
					}
					else {
						t.stop();
						// rkdp is waiting for this before it draws the inner stage itself
						rkdp.stopWaiting();
						repaint();
					}
				}
			});
			// start the animation
			t.start();
		}
		else if (getModel().getState() == State.OUTER) {
			anim = getModel().getStartPoint();
			t = new Timer(1000 / 50, new ActionListener() {
				
				private double d = Math.sqrt((dx(getModel().getEndPoint().getX()) - dx(getModel().getStartPoint().getX())) * (dx(getModel().getEndPoint().getX()) - dx(getModel().getStartPoint().getX())) + (dy(getModel().getEndPoint().getY()) - dy(getModel().getStartPoint().getY())) * (dy(getModel().getEndPoint().getY()) - dy(getModel().getStartPoint().getY())));
				private int nframes = (int) (d * 50 / (getWidth() / 2));
				private int frame = 1;
				
				public void actionPerformed(ActionEvent e) {
					if (frame <= nframes) {
						anim.setLocation(anim.getX() + (getModel().getEndPoint().getX() - getModel().getStartPoint().getX()) / nframes, anim.getY() + (getModel().getEndPoint().getY() - getModel().getStartPoint().getY()) / nframes);
						frame++;
						repaint();
					}
					else {
						t.stop();
						repaint();
					}
				}
			});
			t.start();
		}
		repaint();
	}
}
