/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm;

import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;

import javax.swing.event.ChangeEvent;

/**
 * Panel that provides a geometrical view of the derivatives of the solutions.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RKDerivPanel extends RKDrawing {

	// min and max ordinate in use
	private double minY, maxY;
	
	// whether or not we are waiting for RKAnimationPanel to finish its animation
	private boolean waiting;
	
	/**
	 * Create a new drawing for the given model.
	 */
	public RKDerivPanel(RKModel rkm) {
		super(rkm);
	}
	
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;
		
		if (getModel().getState() != State.EMPTY) {
			drawTitle(getModel().getODE().toString(), g2);
			drawIntervals(getModel().getStartPoint().getX(), getModel().getEndPoint().getX(), minY, maxY, g2);
			plotFunction(getModel().getStartPoint().getX(), getModel().getStartPoint().getY(), true, g2, getPlotColor(), new BasicStroke(1.5f));
		}
		if (getModel().getState() == State.INNER) {
			// draw all but the current stage, this is done without full information
			for (int i = 1; i < getModel().getInnerStagesDone(); i++)
				drawSample(i, false, g2);
			// only draw the current stage if we finished waiting for RKAnimationPanel
			if (! waiting)
				drawSample(getModel().getInnerStagesDone(), true, g2);
		}
		else if (getModel().getState() == State.OUTER) {
			// draw all samples
			for (int i = 1; i <= getModel().getInnerStagesDone(); i++)
				drawSample(i, false, g2);
			// indicate the weighted average by a horizontal line (with tick)
			tickY(getModel().getWeightedSamples(), g2, getOuterStageColor());
			g2.setPaint(getOuterStageColor());
			g2.draw(new Line2D.Float(0, dy(getModel().getWeightedSamples()), getWidth() - 1, dy(getModel().getWeightedSamples())));
		}
	}
	
	// draw an inner stage with or without full information
	private void drawSample(int sample, boolean full, Graphics2D g2) {
		Point2D[] sp = getModel().getSamplePoints();
		double[] s = getModel().getSamples();
		int i = sample - 1;
		
		// only plot the solution derivative if full info is requested
		if (full)
			plotFunction(sp[i].getX(), sp[i].getY(), true, g2, getSecondaryPlotColor(), new BasicStroke());

		tick(sp[i].getX(), s[i], g2);
		
		g2.setPaint(getInnerStagesColor());
		g2.fill(new Ellipse2D.Float(dx(sp[i].getX()) - 3, dy(s[i]) - 3, 6, 6));
	}
	
	/**
	 * Update the view.
	 */
	public void stateChanged(ChangeEvent e) {
		if (getModel().getState() == State.INITIALIZED) {
			// calculate minY and maxY (min/max of samples)
			double[] samples = getModel().getSamples();
			minY = samples[0];
			maxY = samples[0];
			
			for (int i = 1; i < samples.length; i++)
				if (samples[i] < minY)
					minY = samples[i];
				else if (samples[i] > maxY)
					maxY = samples[i];
			
			// we want an interval, even if minY = maxY
			if (minY == maxY) {
				maxY = minY + 0.5;
				minY -= 0.5;
			}
			
			moveToIntervals(getModel().getStartPoint().getX(), getModel().getEndPoint().getX(), minY, maxY);
		}
		// when drawing inner stages, wait for animations to finish first
		else if (getModel().getState() == State.INNER)
			waiting = true;
		repaint();
	}
	
	/**
	 * Inform this panel that animations have finished and the current inner stage should be drawn.
	 */
	public void stopWaiting() {
		waiting = false;
		repaint();
	}
}
