/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.font.TextAttribute;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.text.AttributedString;

import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Superclass for panels that provide a geometrical view.
 * This class provides drawing and plotting functions, manages transformations and colors and handles mouse events.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public abstract class RKDrawing extends JPanel implements ChangeListener {

	// model to listen to
	private RKModel rkm;
	
	// the transformation from true coordinates to screen coordinates is defined by the following parameters
	
	// translation
	private int tx, ty;
	// scaling - the resolutions (pixels per unit) used for each axis
	private double xres, yres;
	
	// colors used in the drawings
	
	// used for coordinates and axes
	private Color coColor;
	// used for the main solution
	private Color plotColor;
	// used for secondary solutions
	private Color secPlotColor;
	// used for inner stages
	private Color isColor;
	// used for the outer stage
	private Color osColor;
	
	/**
	 * Create a new drawing for the given model.
	 */
	public RKDrawing(RKModel rkm) {
		this.rkm = rkm;
		rkm.addChangeListener(this);
		
		// default values
		coColor = Color.BLACK;
		plotColor = Color.BLUE;
		secPlotColor = Color.LIGHT_GRAY;
		isColor = Color.MAGENTA;
		osColor = Color.RED;

		setPreferredSize(new Dimension(400, 400));
		setBackground(Color.WHITE);
		setBorder(BorderFactory.createEtchedBorder());
		
		// catch mouse events
		DrawingMouseListener listener = new DrawingMouseListener();
		addMouseListener(listener);
		addMouseMotionListener(listener);
		addMouseWheelListener(listener);
		
		// rescale the drawing if the user resizes the window
		addComponentListener(new ComponentAdapter() {
			// previous sizes
			private int width = getWidth();
			private int height = getHeight();
			
			// adjust the transformation (scale in the same way the window was scaled) and save sizes for next time
			public void componentResized(ComponentEvent e) {
				if (RKDrawing.this.rkm.getState() != State.EMPTY) {
					xres *= 1.0 * getWidth() / width;
					yres *= 1.0 * getHeight() / height;
					tx *= 1.0 * getWidth() / width;
					ty *= 1.0 * getHeight() / height;
					
					repaint();
				}
				width = getWidth();
				height = getHeight();
			}
		});
	}

	/**
	 * @return the model used.
	 */
	public RKModel getModel() {
		return rkm;
	}
	
	/**
	 * @return the translation used along the X-axis.
	 */
	public int getTranslationX() {
		return tx;
	}
	
	/**
	 * @return the translation used along the Y-axis.
	 */
	public int getTranslationY() {
		return ty;
	}
	
	/**
	 * Set the translations to use.
	 */
	public void setTranslation(int tx, int ty) {
		this.tx = tx;
		this.ty = ty;
		repaint();
	}
	
	/**
	 * @return the resolution used for the X-axis.
	 */
	public double getResolutionX() {
		return xres;
	}
	
	/**
	 * @return the resolution used for the Y-axis.
	 */
	public double getResolutionY() {
		return yres;
	}
	
	/**
	 * Set the resolutions to use.
	 */
	public void setResolution(double xres, double yres) {
		this.xres = xres;
		this.yres = yres;
		repaint();
	}
	
	/**
	 * @return the color used for coordinates and axes.
	 */
	public Color getCoordinateColor() {
		return coColor;
	}
	
	/**
	 * Set the color to use for coordinates and axes.
	 */
	public void setCoordinateColor(Color coColor) {
		this.coColor = coColor;
		repaint();
	}
	
	/**
	 * @return the color used for the main solution.
	 */
	public Color getPlotColor() {
		return plotColor;
	}
	
	/**
	 * Set the color to use for the main solution.
	 */
	public void setPlotColor(Color plotColor) {
		this.plotColor = plotColor;
		repaint();
	}
	
	/**
	 * @return the color used for secondary solutions.
	 */
	public Color getSecondaryPlotColor() {
		return secPlotColor;
	}
	
	/**
	 * Set the color to use for secondary solutions.
	 */
	public void setSecondaryPlotColor(Color secPlotColor) {
		this.secPlotColor = secPlotColor;
		repaint();
	}
	
	/**
	 * @return the color used for inner stages.
	 */
	public Color getInnerStagesColor() {
		return isColor;
	}
	
	/**
	 * Set the color to use for inner stages.
	 */
	public void setInnerStagesColor(Color isColor) {
		this.isColor = isColor;
		repaint();
	}

	/**
	 * @return the color used for the outer stage.
	 */
	public Color getOuterStageColor() {
		return osColor;
	}
	
	/**
	 * Set the color to use for the outer stage.
	 */
	public void setOuterStageColor(Color osColor) {
		this.osColor = osColor;
		repaint();
	}

	// enable antialiasing here so subclasses don't need to do it
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	}
	
	/**
	 * @return the true abscissa for the pixel with screen abscissa x.
	 */
	public double rx(int x) {
		return (x - tx) / xres;
	}
	
	/**
	 * @return the screen abscissa for the point with true abscissa x.
	 */
	public float dx(double x) {
		return (float) (xres * x + tx);
	}
	
	/**
	 * @return the true ordinate for the pixel with screen ordinate y.
	 */
	public double ry(int y) {
		return -(y - ty) / yres;
	}
	
	/**
	 * @return the screen ordinate for the point with true ordinate y.
	 */
	public float dy(double y) {
		return (float) (-yres * y + ty);
	}
	
	/**
	 * Draw the given string in the top left corner.
	 * The coordinate color is used for this.
	 */
	public void drawTitle(String title, Graphics2D g2) {
		g2.setPaint(coColor);
		AttributedString ats = new AttributedString(title);
		ats.addAttribute(TextAttribute.FONT, new Font("Monospaced", Font.BOLD, 12));
		g2.drawString(ats.getIterator(), 4, 12);
	}
	
	/**
	 * Draw axes intervals from minX to maxX and minY to maxY (all true coordinates), with tickmarks at the end points.
	 * The coordinate color is used for this.
	 */
	public void drawIntervals(double minX, double maxX, double minY, double maxY, Graphics2D g2) {
		g2.setPaint(coColor);
		
		g2.draw(new Line2D.Float(dx(minX), (float) (getHeight() * 0.95), dx(maxX), (float) (getHeight() * 0.95)));
		g2.draw(new Line2D.Float((float) (getWidth() * 0.05), dy(maxY), (float) (getWidth() * 0.05), dy(minY)));
		
		tick(minX, minY, g2);
		tick(maxX, maxY, g2);
	}
	
	/**
	 * Put tickmarks on the axes for the point (x, y) (true coordinates).
	 * The coordinate color is used for this.
	 */
	public void tick(double x, double y, Graphics2D g2) {
		tickX(x, g2, coColor);
		tickY(y, g2, coColor);		
	}
	
	/**
	 * Tick the X-axis at point x (true coordinate) using the given color.
	 */
	public void tickX(double x, Graphics2D g2, Color c) {
		g2.setPaint(c);
		g2.draw(new Line2D.Float(dx(x), (float) (getHeight() * 0.95 - 3), dx(x), (float) (getHeight() * 0.95 + 3)));
		g2.drawString(String.format("%.2f", x), dx(x) + 2, (float) (getHeight() * 0.95) + 12);
	}
	
	/**
	 * Tick the Y-axis at point y (true coordinate) using the given color.
	 */
	public void tickY(double y, Graphics2D g2, Color c) {
		g2.setPaint(c);
		g2.draw(new Line2D.Float((float) (getWidth() * 0.05 - 3), dy(y), (float) (getWidth() * 0.05 + 3), dy(y)));
		g2.drawString(String.format("%.2f", y), (float) (getWidth() * 0.05) + 2, dy(y) - 2);
	}
	
	/**
	 * Plot the solution through (x0, y0) (true coordinates) using the given color and stroke.
	 * If deriv is true, the derivate of the solution will be plotted instead of the solution itself.
	 */
	public void plotFunction(double x0, double y0, boolean deriv, Graphics2D g2, Color c, Stroke s) {
		g2.setStroke(s);
		g2.setPaint(c);
		
		// we will plot the function using a GeneralPath made up of line segments
		GeneralPath p = new GeneralPath();
		boolean inDomain = false;
		
		// the following algorithm will only try to plot the part of the function that is visible
		for (int i = 0; i < getWidth(); i ++) {
			// points not in the domain (or near a discontinuity) are skipped
			if (! inDomain) {
				double y = rkm.getODE().evalSol(x0, y0, rx(i));
				while (i < getWidth() - 1 && (Double.isNaN(y) || Double.isInfinite(y) || rkm.getODE().isNearDiscont(x0, y0, rx(i), 1.0 / xres)))
					y = rkm.getODE().evalSol(x0, y0, rx(++i));
				if (i < getWidth()) {
					p.moveTo(i, dy(deriv ? rkm.getODE().evalDeriv(rx(i), y) : y));
					inDomain = true;
				}
			} else {
				double y = rkm.getODE().evalSol(x0, y0, rx(i));
				if (Double.isNaN(y) || Double.isInfinite(y) || rkm.getODE().isNearDiscont(x0, y0, rx(i), 1.0 / xres))
					inDomain = false;
				else
					p.lineTo(i, dy(deriv ? rkm.getODE().evalDeriv(rx(i), y) : y));
			}
		}
		
		g2.draw(p);
		g2.setStroke(new BasicStroke());
	}
	
	/**
	 * Subclasses are required to implement this method.
	 */
	public abstract void stateChanged(ChangeEvent e);

	/**
	 * Scale and translate the view so the given intervals use 70% of the width and height of the panel and are centered.
	 */
	public void moveToIntervals(double minX, double maxX, double minY, double maxY) {
		xres = (0.7 * getWidth()) / (maxX - minX);
		yres = (0.7 * getHeight()) / (maxY - minY);
		tx = (int) Math.floor((getWidth() * 0.15) - (xres * minX));
		ty = (int) Math.floor((getHeight() * 0.15) + (yres * maxY));
	}
	
	// handles mouse events for a panel
	private class DrawingMouseListener extends MouseAdapter implements MouseMotionListener, MouseWheelListener {

		// values recorded when the mouse is pressed
		private int startTx, startTy;
		private double startXres, startYres;
		private int startX, startY;
		
		// whether or not we are scaling (~shift is down as the mouse is pressed)
		private boolean stretching;
		
		// record values and change cursor
		public void mousePressed(MouseEvent e) {
			startTx = tx;
			startTy = ty;
			startXres = xres;
			startYres = yres;
			startX = e.getX();
			startY = e.getY();
			stretching = e.isShiftDown();
			setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
		}

		// return cursor to normal
		public void mouseReleased(MouseEvent e) {
			setCursor(Cursor.getDefaultCursor());
		}

		// move or scale (if shift was down) the drawing
		public void mouseDragged(MouseEvent e) {
			if (e.isShiftDown() && stretching) {
				setResolution(startXres * (e.getX() - tx) / (startX - tx), startYres * (e.getY() - ty) / (startY - ty));
			
				// impose lower and upper bounds
				if (xres < 1)
					xres = 1;
				else if (xres > 10000)
					xres = 10000;
			
				if (yres < 1)
					yres = 1;
				else if (yres > 10000)
					yres = 10000;
			}
			else if (! stretching)
				setTranslation(startTx + (e.getX() - startX), startTy + (e.getY() - startY));
			
			repaint();
		}

		public void mouseMoved(MouseEvent e) {}

		// zoom in or out
		public void mouseWheelMoved(MouseWheelEvent e) {
			double x = rx(e.getX());
			double y = ry(e.getY());
			
			setResolution(xres * (1 - e.getWheelRotation() * 0.05), yres * (1 - e.getWheelRotation() * 0.05));
			
			// impose lower and upper bounds
			if (xres < 1)
				xres = 1;
			else if (xres > 10000)
				xres = 10000;
		
			if (yres < 1)
				yres = 1;
			else if (yres > 10000)
				yres = 10000;
			
			// make sure the same point is still under the mouse after zooming
			setTranslation(tx - (int) (dx(x) - e.getX()), ty - (int) (dy(y) - e.getY()));
			
			repaint();
		}		
	}
}
