/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm;

import java.awt.Component;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.AbstractCellEditor;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTable;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellEditor;
import javax.swing.text.DefaultFormatter;

import rkm.butcher.ButcherMatrix;

/**
 * Dialog window that configures the Runge-Kutta method to use.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RKMethodDialog extends JDialog {

	private RKModel rkm;
	
	// whether or not to use a custom-made Butcher matrix
	private boolean custom;
	// submatrices of the custom Butcher matrix
	private double[][] customA;
	private double[] customB, customC;
	
	private JComboBox BMCombo;
	private JSpinner stagesSpinner;
	private BMTableModel btm;
	private JLabel applyLabel;
	
	private ResourceBundle bundle = ResourceBundle.getBundle("properties/rkm", Locale.getDefault());
	
	/**
	 * Create a new method dialog.
	 */
	public RKMethodDialog(Frame owner, RKModel rkm) {
		super(owner, ResourceBundle.getBundle("properties/rkm", Locale.getDefault()).getString("methodWindowTitle"));
		this.rkm = rkm;
		
		setup();		
		setResizable(false);
	}
	
	// build gui
	private void setup() {
		setContentPane(new JPanel(new GridBagLayout()));
		
		BMCombo = new JComboBox(ButcherMatrix.getAvailableMethods());
		// allow user to select a custom-made Butcher matrix
		BMCombo.addItem(bundle.getString("butcherCustom"));
		BMCombo.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (BMCombo.getSelectedItem().equals(bundle.getString("butcherCustom"))) {
					custom = true;
					stagesSpinner.setEnabled(true);
					createCustomMatrices();
				} else {
					custom = false;
					stagesSpinner.setValue(((ButcherMatrix) BMCombo.getSelectedItem()).order());
					stagesSpinner.setEnabled(false);
				}
				// we need to update both the structure of the table and the data within
				btm.fireTableDataChanged();
				btm.fireTableStructureChanged();
				setApplyReminder();
				pack();
			}
		});
		
		stagesSpinner = new JSpinner(new SpinnerNumberModel(1, 1, 5, 1));
		// only enabled if a Custom is selected
		stagesSpinner.setEnabled(false);
		stagesSpinner.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				createCustomMatrices();
				btm.fireTableDataChanged();
				btm.fireTableStructureChanged();
				setApplyReminder();
				pack();
			}
		});
		
		btm = new BMTableModel();
		JTable btable = new JTable(btm);
		// our custom editor will allow the user to input simple fractions as well
		btable.setDefaultEditor(Double.class, new BMCellEditor());
		
		applyLabel = new JLabel("");
		
		JButton applyButton = new JButton(bundle.getString("buttonApply"));
		applyButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (custom)
					rkm.setButcherMatrix(ButcherMatrix.getCustomButcherMatrix(customA, customB, customC));
				else
					rkm.setButcherMatrix((ButcherMatrix) BMCombo.getSelectedItem());
				applyLabel.setText("");
			}
		});
		
		GridBagConstraints gbc = new GridBagConstraints();
		
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets(5, 0, 0, 5);
		gbc.gridy = 0;
		
		gbc.gridx = 0;
		add(new JLabel(bundle.getString("labelMethod")), gbc);
		gbc.gridx = 1;
		gbc.weightx = 100;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		add(BMCombo, gbc);
		gbc.gridx = 2;
		gbc.weightx = 0;
		gbc.fill = GridBagConstraints.NONE;
		add(new JLabel(bundle.getString("labelStages")), gbc);
		gbc.gridx = 3;
		gbc.insets = new Insets(5, 0, 0, 0);
		add(stagesSpinner, gbc);
		
		gbc.gridy = 1;
		gbc.gridx = 0;
		gbc.gridwidth = 4;
		gbc.weightx = 100;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		add(btable, gbc);
		
		gbc.gridy = 2;
		gbc.gridx = 0;
		gbc.gridwidth = 2;
		gbc.fill = GridBagConstraints.NONE;
		add(applyLabel, gbc);
		
		gbc.gridx = 2;
		gbc.weightx = 0;
		gbc.anchor = GridBagConstraints.EAST;
		add(applyButton, gbc);
		
		pack();
	}
	
	// create new custom matrices bases on the stagesSpinner value
	private void createCustomMatrices() {
		int order = (Integer) stagesSpinner.getValue();
		customA = new double[order][order];
		customB = new double[order];
		customC = new double[order];
	}
	
	// set a label to remind the user to save changes
	private void setApplyReminder() {
		applyLabel.setText(bundle.getString("labelApplyChanges"));
	}
	
	// table model that reads from the Butcher matrix in use
	// will read/edit the custom matrices if building a custom matrix
	private class BMTableModel extends AbstractTableModel {
		public int getColumnCount() {
			if (custom)
				return (Integer) stagesSpinner.getValue() + 1;
			else
				return ((ButcherMatrix) BMCombo.getSelectedItem()).order() + 1;
		}

		public int getRowCount() {
			if (custom)
				return (Integer) stagesSpinner.getValue() + 1;
			else
				return ((ButcherMatrix) BMCombo.getSelectedItem()).order() + 1;
		}

		public Object getValueAt(int rowIndex, int columnIndex) {
			if (custom) {
				int order = (Integer) stagesSpinner.getValue();
				// translate into submatrix indices
				if (columnIndex == 0 && rowIndex == order)
					return null;
				else if (columnIndex == 0)
					return customC[rowIndex];
				else if (rowIndex == order)
					return customB[columnIndex - 1];
				else
					return customA[rowIndex][columnIndex - 1];
			} else {
				if (columnIndex == 0 && rowIndex == ((ButcherMatrix) BMCombo.getSelectedItem()).order())
					return null;
				else if (columnIndex == 0)
					return ((ButcherMatrix) BMCombo.getSelectedItem()).c(rowIndex + 1);
				else if (rowIndex == ((ButcherMatrix) BMCombo.getSelectedItem()).order())
					return ((ButcherMatrix) BMCombo.getSelectedItem()).b(columnIndex);
				else
					return ((ButcherMatrix) BMCombo.getSelectedItem()).a(rowIndex + 1, columnIndex);
			}
		}

		public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
			int order = (Integer) stagesSpinner.getValue();

			if (columnIndex == 0 && rowIndex == order)
				return;
			else if (columnIndex == 0)
				customC[rowIndex] = (Double) aValue;
			else if (rowIndex == order)
				customB[columnIndex - 1] = (Double) aValue;
			else
				customA[rowIndex][columnIndex - 1] = (Double) aValue;
			
			setApplyReminder();
		}

		public boolean isCellEditable(int rowIndex, int columnIndex) {
			if (custom) {
				int order = (Integer) stagesSpinner.getValue();
				// only allow explicit rk-methods (submatrix A has to be subdiagonal)
				return rowIndex >= columnIndex && !(columnIndex == 0 && rowIndex == order);
			} else
				return false;
		}

		public Class<?> getColumnClass(int columnIndex) {
			return Double.class;
		}

		public String getColumnName(int column) {
			return null;
		}
	}
	
	// cell editor that uses a JFormattedTextField as the editor
	// this field will be set up to allow the input of simple fractions
	// this code is similar to the RealFractionFormatter class from the Gaussian Elimination applet
	private class BMCellEditor extends AbstractCellEditor implements TableCellEditor {

		// field to use as the editor
		private JFormattedTextField field;
		
		public BMCellEditor() {
			// create the field using a custom formatter that allows fractions
			field = new JFormattedTextField(new DefaultFormatter() {				
				// will parse the numerator and denominator
				private NumberFormat nf = NumberFormat.getNumberInstance();
				
				public Object stringToValue(String string) throws ParseException {
					int sign = string.indexOf('/');

					try {
						if (sign == -1)
							return nf.parse(string).doubleValue();
						else {
							double numerator = nf.parse(string.substring(0, sign)).doubleValue();
							double denominator = nf.parse(string.substring(sign + 1)).doubleValue();
							
							if (denominator != 0)
								return numerator / denominator;
							else
								return numerator;
						}
					} catch (NumberFormatException e) {
						throw new ParseException("Cannot parse this value.", 297);
					}
				}

				public String valueToString(Object value) throws ParseException {
					return value == null ? bundle.getString("rffUnknownValue") : value.toString();
				}
			});
			
			// finish editing when the user presses enter
			field.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					stopCellEditing();
				}
			});
		}
		
		// return the formatted text field as the editor
		public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row, int column) {
			field.setValue(value);
			return field;
		}

		public Object getCellEditorValue() {
			return field.getValue();
		}
	}
}
