/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm;

import java.awt.geom.Point2D;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import rkm.butcher.ButcherMatrix;
import rkm.butcher.ForwardEuler;
import rkm.ode.ExampleODE1;
import rkm.ode.ODE;

/**
 * The state of the model.
 * Possible values are:
 * <ul>
 * <li>EMPTY: not initialized yet.</li>
 * <li>INITIALIZED: initialized but no stages done yet.</li>
 * <li>INNER: one or more inner stages done.</li>
 * <li>OUTER: outer stage done (algorithm is finished).</li>
 * </ul>
 */
enum State {EMPTY, INITIALIZED, INNER, OUTER};

/**
 * Runge-Kutta algorithm for solving initial value problems.
 * This class takes care of all calculations and remembers the state of the model.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RKModel {

	// the ODE to solve
	private ODE ode;
	// RK method to use
	private ButcherMatrix bm;

	// current state
	private State state;
	// count of inner stages done
	private int stagesDone;
	
	// start point (initial values)
	private Point2D start;
	// end point (approximated solution)
	private Point2D end;
	// step size
	private double step;
	// absolute error
	private double error;
	
	// sample points (where the slopes are calculated)
	private Point2D[] samplePoints;
	// samples (calculated slopes)
	private double[] samples;
	// weighted average of samples
	private double weightedSamples;
	
	// listeners
	private EventListenerList ll;
	
	/**
	 * Create a new model to solve the given ODE using the given ButcherMatrix.
	 * The default ODE is ExampleODE1, the default ButcherMatrix is ForwardEuler.
	 */
	public RKModel(ODE ode, ButcherMatrix bm) {
		this.ode = (ode == null ? new ExampleODE1() : ode);
		this.bm = (bm == null ? new ForwardEuler() : bm);
		
		empty();
			
		ll = new EventListenerList();
	}
	
	// resets the state variables
	// this does not clear any results so don't retrieve them when the state is EMPTY
	private void empty() {
		state = State.EMPTY;
		stagesDone = 0;
	}
	
	/**
	 * @return the ODE that is being solved.
	 */
	public ODE getODE() {
		return ode;
	}

	/**
	 * Set the ODE to solve. This will reset the model. 
	 * The default ODE is ExampleODE1 (used if the argument is null).
	 */
	public void setODE(ODE ode) {
		this.ode = (ode == null ? new ExampleODE1() : ode);
		empty();
		fireStateChanged(new ChangeEvent(this));
	}
	
	/**
	 * @return the ButcherMatrix that is in use.
	 */
	public ButcherMatrix getButcherMatrix() {
		return bm;
	}

	/**
	 * Set the ButcherMatrix to use. This will reset the model.
	 * The default ButcherMatrix is ForwardEuler (used if the argument is null).
	 */
	public void setButcherMatrix(ButcherMatrix bm) {
		this.bm = (bm == null ? new ForwardEuler() : bm);
		empty();
		fireStateChanged(new ChangeEvent(this));
	}
	
	/**
	 * Initialize the algoritm with the given initial values.
	 * After this step (if successful) all results are available, regardless of the state of the model.
	 * 
	 * @return false if the given values are invalid or cannot be used.
	 */
	public boolean init(double x0, double y0, double h) {
		boolean valid = ode.isValidInterval(x0, y0, h);
		if (valid) {
			start = new Point2D.Double(x0, y0);
			step = h;
			state = State.INITIALIZED;
			stagesDone = 0;
			step(x0, y0, h);
			fireStateChanged(new ChangeEvent(this));
		}
		return valid;
	}
	
	// compute the results with the runge-kutta algorithm
	// in case you get confused by the indices: the ButcherMatrix public methods use indices 1...n (not 0...n-1)
	private void step(double x0, double y0, double h) {
		samples = new double[bm.order()];
		samplePoints = new Point2D[bm.order()];
		weightedSamples = 0;
		for (int i = 1; i <= bm.order(); i++) {
			double wa = 0;
			for (int j = 1; j < i; j++) 
				wa += bm.a(i, j) * samples[j - 1];
			samplePoints[i - 1] = new Point2D.Double(x0 + bm.c(i) * h, y0 + h * wa);
			samples[i - 1] = ode.evalDeriv(samplePoints[i - 1].getX(), samplePoints[i - 1].getY());
			weightedSamples += bm.b(i) * samples[i - 1];
		}
		end = new Point2D.Double(x0 + h, y0 + h * weightedSamples);
		error = end.getY() - ode.evalSol(x0, y0, x0 + h);
	}
	
	/**
	 * @return whether there are inner stages remaining.
	 */
	public boolean hasInnerStagesRemaining() {
		return state == State.INITIALIZED || state == State.INNER && stagesDone < bm.order();
	}
	
	/**
	 * Sets the state of the model one inner stage further.
	 */
	public void nextInnerStage() {
		if (state == State.INITIALIZED) {
			stagesDone = 1;
			state = State.INNER;
			fireStateChanged(new ChangeEvent(this));
		} else if (state == State.INNER && stagesDone < bm.order()) {
			stagesDone++;
			fireStateChanged(new ChangeEvent(this));
		}
	}
	
	/**
	 * Sets the state of the model to indicate that the outer stage is done.
	 */
	public void outerStage() {
		if (state == State.INNER && stagesDone == bm.order()) {
			state = State.OUTER;
			fireStateChanged(new ChangeEvent(this));
		}
	}
	
	/**
	 * @return the state of the model.
	 */
	public State getState() {
		return state;
	}

	/**
	 * @return the number of inner stages done.
	 */
	public int getInnerStagesDone() {
		return stagesDone;
	}
	
	/**
	 * @return the initial values used as a starting point.
	 */
	public Point2D getStartPoint() {
		return (Point2D) start.clone();
	}
	
	/**
	 * @return the approximated solution point.
	 */
	public Point2D getEndPoint() {
		return (Point2D) end.clone();
	}

	/**
	 * @return the step size used.
	 */
	public double getStepSize() {
		return step;
	}
	
	/**
	 * @return the absolute error in the approximated solution point.
	 */
	public double getError() {
		return error;
	}
	
	/**
	 * @return the points in which the slope samples were calculated.
	 */
	public Point2D[] getSamplePoints() {
		Point2D[] clone = new Point2D[samplePoints.length];
		for (int i = 0; i < samplePoints.length; i++)
			clone[i] = (Point2D) samplePoints[i].clone();
		return clone;
	}
	
	/**
	 * @return the slope samples.
	 */
	public double[] getSamples() {
		return samples.clone();
	}
	
	/**
	 * @return a weighted average of slope samples.
	 */
	public double getWeightedSamples() {
		return weightedSamples;
	}
	
	/**
	 * Add a ChangeListener.
	 */
	public void addChangeListener(ChangeListener l) {
		ll.add(ChangeListener.class, l);
	}

	/**
	 * Remove a ChangeListener.
	 */
	public void removeChangeListener(ChangeListener l) {
		ll.remove(ChangeListener.class, l);
	}

	/**
	 * Notify all ChangeListeners.
	 */
	public void fireStateChanged(ChangeEvent e) {
		ChangeListener[] listeners = ll.getListeners(ChangeListener.class);
		for (ChangeListener l : listeners)
			l.stateChanged(e);
	}
}