/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;

/**
 * Panel that displays numerical results based on the state of the model.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RKNumPanel extends JPanel implements ChangeListener {

	// model to listen to
	private RKModel rkm;
	
	// table model for inner stages data
	private NumTableModel numModel;
	// weighted average of samples
	private JLabel waValue;
	// approximated solution point
	private JLabel osValues;
	// absolute error in the approximated solution point
	private JLabel errorValue;
	
	private ResourceBundle bundle = ResourceBundle.getBundle("properties/rkm", Locale.getDefault());
	
	/**
	 * Create a new panel that listens to the given model.
	 */
	public RKNumPanel(RKModel rkm) {
		super(new GridBagLayout());
		this.rkm = rkm;
		rkm.addChangeListener(this);
		
		setup();
	}
	
	// build gui
	private void setup() {
		setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(bundle.getString("borderNumResults")), BorderFactory.createEmptyBorder(5, 5, 5, 5)));
		
		numModel = new NumTableModel();
		JTable numTable = new JTable(numModel);
		JPanel numTablePanel = new JPanel(new BorderLayout());
		
		waValue = new JLabel("...");
		JPanel waPanel = new JPanel();
		waPanel.add(waValue);
		
		// if we don't put it into a panel, the table will be displayed without a heading
		numTablePanel.add(numTable.getTableHeader(), BorderLayout.NORTH);
		numTablePanel.add(numTable, BorderLayout.CENTER);
		numTablePanel.add(waPanel, BorderLayout.SOUTH);
		
		osValues = new JLabel("...");
		errorValue = new JLabel("...");
		
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weightx = 100;
		gbc.gridx = 0;
		gbc.insets = new Insets(5, 0, 0, 0);
		
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weighty = 0;
		
		gbc.gridy = 0;
		add(new JLabel(bundle.getString("labelInnerStages")), gbc);
		gbc.gridy = 2;
		add(new JLabel(bundle.getString("labelOuterStage")), gbc);
		gbc.gridy = 4;
		add(new JLabel(bundle.getString("labelError")), gbc);
		
		gbc.anchor = GridBagConstraints.CENTER;
				
		gbc.gridy = 1;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.weighty = 100;
		add(numTablePanel, gbc);
		gbc.gridy = 3;
		gbc.fill = GridBagConstraints.NONE;
		gbc.weighty = 0;
		add(osValues, gbc);
		gbc.gridy = 5;
		add(errorValue, gbc);
	}
	
	/**
	 * Update the panel.
	 */
	public void stateChanged(ChangeEvent e) {
		// update inner stages data
		numModel.fireTableDataChanged();
		
		// update labels
		if (rkm.getState() == State.OUTER) {
			waValue.setText(bundle.getString("labelAverage") + " " + String.format("%.3f", rkm.getWeightedSamples()));
			osValues.setText(String.format("( %.3f , %.3f ) -> ( %.3f , %.3f )", rkm.getStartPoint().getX(), rkm.getStartPoint().getY(), rkm.getEndPoint().getX(), rkm.getEndPoint().getY()));
			errorValue.setText("" + rkm.getError());
		} else {
			waValue.setText("...");
			osValues.setText("...");
			errorValue.setText("...");
		}
	}
	
	// table model for inner stages data
	private class NumTableModel extends AbstractTableModel {

		public int getColumnCount() {
			return 4;
		}

		// row count depends on the rk-method used
		public int getRowCount() {
			return rkm.getButcherMatrix().order();
		}

		// only return values for inner stages that are done, otherwise return null
		public Object getValueAt(int rowIndex, int columnIndex) {
			if (columnIndex == 0 && rowIndex < rkm.getInnerStagesDone())
				return rowIndex + 1;
			else if (columnIndex == 1 && rowIndex < rkm.getInnerStagesDone())
				return rkm.getSamplePoints()[rowIndex].getX();
			else if (columnIndex == 2 && rowIndex < rkm.getInnerStagesDone())
				return rkm.getSamplePoints()[rowIndex].getY();
			else if (columnIndex == 3 && rowIndex < rkm.getInnerStagesDone())
				return rkm.getSamples()[rowIndex];
			else
				return null;				
		}

		public Class<?> getColumnClass(int columnIndex) {
			if (columnIndex == 0)
				return Integer.class;
			else
				return Double.class;
		}

		public String getColumnName(int column) {
			return new String[] {bundle.getString("labelStageColumn"), "x", "y", "k"}[column];
		}
	}
}