/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.NumberFormat;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import rkm.ode.ODE;

/**
 * Panel that controls the IVP to solve and the state of the model.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RKProblemPanel extends JPanel implements ChangeListener {

	private RKModel rkm;
	
	// selection of ODEs
	private JComboBox ODECombo;
	// input values
	private JFormattedTextField pField, xField, yField, hField;
	// controller buttons
	private JButton startButton, isButton, osButton;	
	
	private ResourceBundle bundle = ResourceBundle.getBundle("properties/rkm", Locale.getDefault());
	
	/**
	 * Create a new problem panel.
	 */
	public RKProblemPanel(RKModel rkm) {
		super(new GridBagLayout());
		this.rkm = rkm;
		rkm.addChangeListener(this);
		
		setup();
	}
	
	// build gui
	private void setup() {
		setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(bundle.getString("borderProblem")), BorderFactory.createEmptyBorder(5, 5, 5, 5)));

		ODECombo = new JComboBox(ODE.getAvailableODEs());
		ODECombo.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// set the ODE
				rkm.setODE((ODE) ODECombo.getSelectedItem());
				
				// see if the new ODE requires a parameter and set the input values to what the ODE suggests
				pField.setEnabled(rkm.getODE().requiresParameter());
				xField.setValue(rkm.getODE().getStartPointSuggestion().getX());
				yField.setValue(rkm.getODE().getStartPointSuggestion().getY());
				hField.setValue(rkm.getODE().getStepSizeSuggestion());
			}
		});
		
		pField = new JFormattedTextField(NumberFormat.getNumberInstance());
		xField = new JFormattedTextField(NumberFormat.getNumberInstance());
		yField = new JFormattedTextField(NumberFormat.getNumberInstance());
		hField = new JFormattedTextField(NumberFormat.getNumberInstance());
		
		pField.setEnabled(rkm.getODE().requiresParameter());
		xField.setValue(rkm.getODE().getStartPointSuggestion().getX());
		yField.setValue(rkm.getODE().getStartPointSuggestion().getY());
		hField.setValue(rkm.getODE().getStepSizeSuggestion());
		
		startButton = new JButton(bundle.getString("buttonStart"));
		isButton = new JButton(bundle.getString("buttonInnerStage"));
		osButton = new JButton(bundle.getString("buttonOuterStage"));
		
		isButton.setEnabled(false);
		osButton.setEnabled(false);
		
		startButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				Number p = (Number) pField.getValue();
				Number x = (Number) xField.getValue();
				Number y = (Number) yField.getValue();
				Number step = (Number) hField.getValue();
				
				// make sure a parameter is supplied if one is needed
				if (rkm.getODE().requiresParameter()) {
					if (p == null || p.doubleValue() == 0) {
						JOptionPane.showMessageDialog(startButton, bundle.getString("missingParameterMessage"), bundle.getString("missingParameterTitle"), JOptionPane.ERROR_MESSAGE);
						return;
					}
					else
						rkm.getODE().setParameter(p.doubleValue());
				}
				
				// see if the input values are valid, and if so initialize the model
				if (x != null && y != null && step != null && rkm.init(x.doubleValue(), y.doubleValue(), step.doubleValue())) {
					isButton.setEnabled(true);
				} else
					JOptionPane.showMessageDialog(startButton, bundle.getString("invalidParameterMessage"), bundle.getString("invalidParameterTitle"), JOptionPane.ERROR_MESSAGE);
			}
		});
		
		isButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				rkm.nextInnerStage();
			}
		});
		
		osButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				rkm.outerStage();
			}
		});
		
		JPanel buttonsPanel = new JPanel(new GridLayout(1, 3, 5, 0));
		buttonsPanel.add(startButton);
		buttonsPanel.add(isButton);
		buttonsPanel.add(osButton);
		
		GridBagConstraints gbc = new GridBagConstraints();
		
		gbc.weighty = 100 / 6;
		
		gbc.anchor = GridBagConstraints.WEST;		
		gbc.gridx = 0;
		gbc.weightx = 20;
		
		gbc.gridy = 0;
		add(new JLabel(bundle.getString("labelEquation")), gbc);
		gbc.gridy = 1;
		gbc.insets = new Insets(5, 0, 0, 10);
		add(new JLabel(bundle.getString("labelParameter")), gbc);
		gbc.gridy = 2;
		add(new JLabel("<html>x<sub>0</sub>: </html>"), gbc);
		gbc.gridy = 3;
		add(new JLabel("<html>y<sub>0</sub>: </html>"), gbc);
		gbc.gridy = 4;
		add(new JLabel(bundle.getString("labelStepSize")), gbc);
		
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.gridx = 1;
		gbc.weightx = 80;
		
		gbc.gridy = 0;
		gbc.insets = new Insets(0, 0, 0, 0);
		add(ODECombo, gbc);
		gbc.gridy = 1;
		gbc.insets = new Insets(5, 0, 0, 0);
		add(pField, gbc);
		gbc.gridy = 2;
		add(xField, gbc);
		gbc.gridy = 3;
		add(yField, gbc);
		gbc.gridy = 4;
		add(hField, gbc);
		
		gbc.gridx = 0;
		gbc.gridwidth = 2;
		gbc.gridy = 5;
		add(buttonsPanel, gbc);
	}

	public void stateChanged(ChangeEvent e) {
		isButton.setEnabled(rkm.getState() == State.INITIALIZED || rkm.getState() == State.INNER && rkm.hasInnerStagesRemaining());
		osButton.setEnabled(rkm.getState() == State.INNER && ! rkm.hasInnerStagesRemaining());
	}
}
