/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm.butcher;

/**
 * Butcher matrix representation of a Runge-Kutta method.
 * This class also keeps track of all available subclasses.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public abstract class ButcherMatrix {

	// submatrix A
	protected double[][] a;
	// submatrix B
	protected double[] b;
	// submatrix C
	protected double[] c;
	
	/**
	 * @return element (i,j) of submatrix A.
	 * Both indices range from 1 to order.
	 */
	public double a(int i, int j) {
		return a[i - 1][j - 1];
	}

	/**
	 * @return element i of submatrix B.
	 * The range of i is from 1 to order.
	 */
	public double b(int i) {
		return b[i - 1];
	}
	
	/**
	 * @return element i of submatrix C.
	 * The range of i is from 1 to order.
	 */
	public double c(int i) {
		return c[i - 1];
	}
	
	/**
	 * @return the order of the matrix.
	 * This is the size (n) of submatrices A (nxn), B (1xn) and C (nx1).
	 */
	public abstract int order();
	
	/**
	 * @return a description of the Runge-Kutta method this matrix represents.
	 */
	public abstract String toString();
	
	
	/**
	 * @return an array of ButcherMatrices, one of each available subclass.
	 */
	public static ButcherMatrix[] getAvailableMethods() {
		return new ButcherMatrix[] {new ForwardEuler(), new Midpoint(), new ModifiedEuler(), new Heun(), new RungeKutta3(), new RungeKutta4(), new ThreeEights(), new Merson()};
	}
	
	/**
	 * @return a custom-built Butcher matrix built with the given submatrices.
	 */
	public static ButcherMatrix getCustomButcherMatrix(double[][] a, double[] b, double[] c) {
		return new CustomButcherMatrix(a, b, c);
	}
}
