/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package rkm.ode;

import java.awt.geom.Point2D;

/**
 * Information about an ODE and its solutions.
 * This class also keeps track of all available subclasses.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public abstract class ODE {
	
	// parameter (if used in the ODE)
	private double p;
	
	/**
	 * @return whether this ODE uses a parameter.
	 */
	public abstract boolean requiresParameter();
	
	/**
	 * @return the parameter used (if used).
	 */
	public double getParameter() {
		return p;
	}
	
	/**
	 * Set the parameter to use.
	 * ODEs that don't require a parameter will ignore this.
	 */
	public void setParameter(double p) {
		this.p = p;
	}	
	
	/**
	 * Evaluate this ODE at point (x,y).
	 */
	public abstract double evalDeriv(double x, double y);
	
	/**
	 * Evaluate the solution through (x0,y0) at point x.
	 */
	public abstract double evalSol(double x0, double y0, double x);
	
	/**
	 * @return whether the solution through (x0,y0) at point x is within distance eps of a discontinuity.
	 * This information is used when plotting the solution.
	 */
	public abstract boolean isNearDiscont(double x0, double y0, double x, double eps);
	
	/**
	 * @return whether the given values can be used as initial values to solve this ODE.
	 */
	public abstract boolean isValidInterval(double x0, double y0, double width);
	
	/**
	 * @return a suggested starting point.
	 */
	public abstract Point2D getStartPointSuggestion();
	
	/**
	 * @return a suggested step size.
	 */
	public abstract double getStepSizeSuggestion();
	
	/**
	 * @return a description of this ODE.
	 */
	public abstract String toString();
	
	/**
	 * @return a description of the solutions for this ODE.
	 */
	public abstract String getSolutionDescription();
	
	/**
	 * @return an array of ODEs, one of each available subclass.
	 */
	public static ODE[] getAvailableODEs() {
		return new ODE[] {new ExampleODE1(), new ExampleODE2(), new ExampleODE3(), new ExampleODE4(), new ExampleODE5(), new ExampleODE6(), new ExampleODE7()};
	}
}
